/*
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of I.B.M. not be used in advertising
// or publicity pertaining to distribution of the software without specific,
// written prior permission. I.B.M. makes no representations about the
// suitability of this software for any purpose.  It is provided "as is"
// without express or implied warranty.
//
// I.B.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL I.B.M.
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Author:  Marc Andreessen, Barry Minor, IBM AWS Graphics Systems (Austin)
// Special Thanks to the Men and Women of DEC ...
*/

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>

static struct tms tbuf;
static int gtime;
FILE *outfile;

/* Lifted from x11perf; round to 3 digits. */

static float 
roundit (float d)
{
  float exponent;
  
  exponent = 1.0;
  if (d >= 1000.0) {
    do {
      exponent *= 10.0;
    } while (d/exponent >= 1000.0);
    d = (float)((int) (d/exponent + 0.5));
    d *= exponent;
  } else {
    while (d*exponent < 100.0) {
      exponent *= 10.0;
    }
    d = (float)((int) (d*exponent + 0.5));
    d /= exponent;
  }
  return d;
}


void startclock (void) 
{
  glFinish ();
  gtime = times(&tbuf);
  
  return;
}

void stopclock (int frames, char *objname)
{
  float period;
  float fps;
  float timeperstruct;
  float rate, crate;
  int quiet = 0;
  
  glFinish ();
  period = (float)(times(&tbuf) - gtime) / 100.0;
  fps = (float)frames/period;

  fprintf(stdout,"%3.2f\tframes/sec\t--\t%s\n", fps, objname);

  return;
}

