/*
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of I.B.M. not be used in advertising
// or publicity pertaining to distribution of the software without specific,
// written prior permission. I.B.M. makes no representations about the
// suitability of this software for any purpose.  It is provided "as is"
// without express or implied warranty.
//
// I.B.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL I.B.M.
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Author:  Barry Minor, IBM AWS Graphics Systems (Austin)
*/

#define PRP__setup							\
        static float cycle=PI;						\
        static float yang;						\
        yang = cos(cycle)+1.0;

#define LIT__setup							\
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);		\
	glEnable(GL_COLOR_MATERIAL);

#define LIT__reset							\
	glDisable(GL_COLOR_MATERIAL);

#define MULTI_PLAIN_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
glBegin(##bgnclass##);					\
outer##;							\
for(j=0;j<ply[i].numverts;++j)					\
{								\
inner1##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
}								\
glEnd();							\
}

#define MESH_PLAIN_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)     \
{                                                               \
glBegin(##bgnclass##);                                      \
outer##;                                                      \
for(j=0;j<msh[i].numverts;++j)                                  \
{                                                               \
inner1##;                                                     \
glVertex3fv(&msh[i].verts[j].x);                                     \
}                                                               \
glEnd();                                                        \
}

#define MULTI_PRP_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
glPushMatrix();							\
glRotatef(((float) i)*yang, 0.0, 1.0, 0.0);				\
glBegin(##bgnclass##);					\
outer##;							\
for(j=0;j<ply[i].numverts;++j)					\
{								\
inner1##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
}								\
glEnd();							\
glPopMatrix();							\
}								\
cycle += PI/30.0;

#define MESH_PRP_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)       \
{                                                               \
glPushMatrix();                                                 \
glRotatef(((float) i)*yang, 0.0, 1.0, 0.0);                             \
glBegin(##bgnclass##);                                      \
outer##;                                                      \
for(j=0;j<msh[i].numverts;++j)                                  \
{                                                               \
inner1##;                                                     \
glVertex3fv(&msh[i].verts[j].x);                                     \
}                                                               \
glEnd();                                                        \
glPopMatrix();                                                  \
}                                                               \
cycle += PI/30.0;

#define SINGLE_PLAIN_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
outer##;							\
for(j=0;j<ply[i].numverts-1;++j)				\
{								\
glBegin(##bgnclass##);					\
inner1##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
inner2##;							\
glVertex3fv(vert[ply[i].index[j+1]]);					\
glEnd();							\
}								\
glBegin(##bgnclass##);					\
inner3##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
inner4##;							\
glVertex3fv(vert[ply[i].index[0]]);					\
glEnd();							\
}

#define SINGLE_MESH_PLAIN_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
outer##;							\
for(j=0;j<msh[i].numverts-1;++j)                                  \
{								\
glBegin(##bgnclass##);					\
inner1##;							\
glVertex3fv(&msh[i].verts[j].x);                                     \
inner2##;							\
glVertex3fv(&msh[i].verts[j+1].x);                                     \
glEnd();							\
}								\
}

#define SINGLE_PRP_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
glPushMatrix();							\
glRotatef(((float) i)*yang, 0.0, 1.0, 0.0);				\
outer##;							\
for(j=0;j<ply[i].numverts-1;++j)				\
{								\
glBegin(##bgnclass##);					\
inner1##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
inner2##;							\
glVertex3fv(vert[ply[i].index[j+1]]);					\
glEnd();							\
}								\
glBegin(##bgnclass##);					\
inner3##;							\
glVertex3fv(vert[ply[i].index[j]]);					\
inner4##;							\
glVertex3fv(vert[ply[i].index[0]]);					\
glEnd();							\
glPopMatrix();							\
}								\
cycle += PI/30.0;

#define SINGLE_MESH_PRP_LOOP(bgnclass,outer,inner1,inner2,inner3,inner4)	\
{								\
glPushMatrix();							\
glRotatef(((float) i)*yang, 0.0, 1.0, 0.0);				\
outer##;							\
for(j=0;j<msh[i].numverts-1;++j)                                  \
{								\
glBegin(##bgnclass##);					\
inner1##;							\
glVertex3fv(&msh[i].verts[j].x);                                     \
inner2##;							\
glVertex3fv(&msh[i].verts[j+1].x);                                     \
glEnd();							\
}								\
glPopMatrix();							\
}								\
cycle += PI/30.0;

#define GEN_PLAIN(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,;,;,;,;)			\
}

#define GEN_FC(loopclass,bgnclass,rootname)					\
void rootname()								\
{									\
int i,j;								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor,;,;,;,;)		\
}

#define GEN_VC(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innercolor,innercolor12,innercolor,innercolor22)					\
}

#define GEN_PRP(loopclass,bgnclass,rootname)					\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,;,;,;,;)						\
}

#define GEN_TEXTURE(loopclass,bgnclass,rootname)					\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innertex,innertex12,innertex,innertex22)					\
}


#define GEN_FC_PRP(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor,;,;,;,;)			\
}

#define GEN_VC_PRP(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innercolor,innercolor12,innercolor,innercolor22)			\
}


#define GEN_FC_TEXTURE(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor,innertex,innertex12,innertex,innertex22)						\
}

#define GEN_VC_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innercolor;innertex,innercolor12;innertex12,innercolor;innertex,innercolor22;innertex22)						\
}

#define GEN_PRP_TEXTURE(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innertex,innertex12,innertex,innertex22)					\
}

#define GEN_FC_PRP_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor,innertex,innertex12,innertex,innertex22)				\
}

#define GEN_VC_PRP_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innercolor;innertex,innercolor12;innertex12,innercolor;innertex,innercolor22;innertex22);		\
}


#define GEN_VN(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innernorm,innernorm12,innernorm,innernorm22)					\
}

#define GEN_VN_FC(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor,innernorm,innernorm12,innernorm,innernorm22)				\
LIT__reset								\
}

#define GEN_VN_VC(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innercolor;innernorm,innercolor12;innernorm12,innercolor;innernorm,innercolor22;innernorm22)		\
LIT__reset								\
}

#define GEN_VN_PRP(loopclass,bgnclass,rootname)					\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innernorm,innernorm12,innernorm,innernorm22)					\
}

#define GEN_VN_TEXTURE(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innertex;innernorm,innertex12;innernorm12,innertex;innernorm,innertex22;innernorm22)	\
}


#define GEN_VN_FC_PRP(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor,innernorm,innernorm12,innernorm,innernorm22)	\
LIT__reset								\
}

#define GEN_VN_VC_PRP(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innercolor;innernorm,innercolor12;innernorm12,innercolor;innernorm,innercolor22;innernorm22)			\
LIT__reset								\
}


#define GEN_VN_FC_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor,innertex;innernorm,innertex12;innernorm12,innertex;innernorm,innertex22;innernorm22)	\
LIT__reset								\
}

#define GEN_VN_VC_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,;,innercolor;innertex;innernorm,innercolor12;innertex12;innernorm12,innercolor;innertex;innernorm,innercolor22;innertex22;innernorm22)\
LIT__reset								\
}

#define GEN_VN_PRP_TEXTURE(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innertex;innernorm,innertex12;innernorm12,innertex;innernorm,innertex22;innernorm22)		\
}

#define GEN_VN_FC_PRP_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor,innertex;innernorm,innertex12;innernorm12,innertex;innernorm,innertex22;innernorm22)	\
LIT__reset								\
}

#define GEN_VN_VC_PRP_TEXTURE(loopclass,bgnclass,rootname)		\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,;,innercolor;innertex;innernorm,innercolor12;innertex12;innernorm12,innercolor;innertex;innernorm,innercolor22;innertex22;innernorm22)\
LIT__reset								\
}


#define GEN_FN(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outernormal,;,;,;,;)					\
}

#define GEN_FN_FC(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor;outernormal,;,;,;,;)		\
LIT__reset								\
}

#define GEN_FN_VC(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outernormal,innercolor,innercolor12,innercolor,innercolor22)				\
LIT__reset								\
}

#define GEN_FN_PRP(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outernormal,;,;,;,;)			\
}

#define GEN_FN_TEXTURE(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outernormal,innertex,innertex12,innertex,innertex22)				\
}


#define GEN_FN_FC_PRP(loopclass,bgnclass,rootname)				\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor;outernormal,;,;,;,;)			\
LIT__reset								\
}

#define GEN_FN_VC_PRP(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outernormal,innercolor,innercolor12,innercolor,innercolor22)		\
LIT__reset								\
}


#define GEN_FN_FC_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outercolor;outernormal,innertex,innertex12,innertex,innertex22)	\
LIT__reset								\
}

#define GEN_FN_VC_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PLAIN_LOOP(bgnclass,outernormal,innercolor;innertex,innercolor12;innertex12,innercolor;innertex,innercolor22;innertex22)	\
LIT__reset								\
}

#define GEN_FN_PRP_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
glColor3f(1.0,1.0,1.0);							\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outernormal,innertex,innertex12,innertex,innertex22)				\
}

#define GEN_FN_FC_PRP_TEXTURE(loopclass,bgnclass,rootname)			\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outercolor;outernormal,innertex,innertex12,innertex,innertex22)		\
LIT__reset								\
}

#define GEN_FN_VC_PRP_TEXTURE(loopclass,bgnclass,rootname)		\
void rootname()								\
{									\
int i,j;								\
PRP__setup								\
LIT__setup								\
for(i=0;i<np;++i)							\
loopclass##PRP_LOOP(bgnclass,outernormal,innercolor;innertex,innercolor12;innertex12,innercolor;innertex,innercolor22;innertex22)\
LIT__reset								\
}

#define EXPAND_TEXTURE(loopclass,gclass,bgnclass,rootname)              \
        gclass##_TEXTURE(loopclass,bgnclass,rootname##texture)

#define EXPAND_PRP(loopclass,gclass,bgnclass,rootname)                  \
        EXPAND_TEXTURE(loopclass,gclass##_PRP,bgnclass,rootname##prp)\
        gclass##_PRP(loopclass,bgnclass,rootname##prp)

#define EXPAND_VC(loopclass,gclass,bgnclass,rootname)           \
        EXPAND_PRP(loopclass,gclass##_VC,bgnclass,rootname##vertcolor)\
        EXPAND_TEXTURE(loopclass,gclass##_VC,bgnclass,rootname##vertcolor)\
        gclass##_VC(loopclass,bgnclass,rootname##vertcolor)

#define EXPAND_FC(loopclass,gclass,bgnclass,rootname)                   \
        EXPAND_PRP(loopclass,gclass##_FC,bgnclass,rootname##color)  \
        EXPAND_TEXTURE(loopclass,gclass##_FC,bgnclass,rootname##color) \
        gclass##_FC(loopclass,bgnclass,rootname##color)

#define EXPAND_FN(loopclass,gclass,bgnclass,rootname) \
        EXPAND_FC(loopclass,gclass##_FN,bgnclass,rootname##fn) \
        EXPAND_VC(loopclass,gclass##_FN,bgnclass,rootname##fn)      \
        EXPAND_PRP(loopclass,gclass##_FN,bgnclass,rootname##fn) \
        EXPAND_TEXTURE(loopclass,gclass##_FN,bgnclass,rootname##fn) \
        gclass##_FN(loopclass,bgnclass,rootname##fn)

#define EXPAND_VN(loopclass,gclass,bgnclass,rootname)           \
        EXPAND_FC(loopclass,gclass##_VN,bgnclass,rootname##vn) \
        EXPAND_VC(loopclass,gclass##_VN,bgnclass,rootname##vn)      \
        EXPAND_PRP(loopclass,gclass##_VN,bgnclass,rootname##vn) \
        EXPAND_TEXTURE(loopclass,gclass##_VN,bgnclass,rootname##vn) \
        gclass##_VN(loopclass,bgnclass,rootname##vn)

#define EXPAND(loopclass,bgnclass,rootname)                             \
        EXPAND_FC(loopclass,GEN,bgnclass,rootname)                      \
        EXPAND_VC(loopclass,GEN,bgnclass,rootname)              \
        EXPAND_PRP(loopclass,GEN,bgnclass,rootname)                     \
        EXPAND_TEXTURE(loopclass,GEN,bgnclass,rootname)                 \
        EXPAND_VN(loopclass,GEN,bgnclass,rootname)                      \
        EXPAND_FN(loopclass,GEN,bgnclass,rootname)                      \
        GEN_PLAIN(loopclass,bgnclass,rootname)

