/*
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of I.B.M. not be used in advertising
// or publicity pertaining to distribution of the software without specific,
// written prior permission. I.B.M. makes no representations about the
// suitability of this software for any purpose.  It is provided "as is"
// without express or implied warranty.
//
// I.B.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL I.B.M.
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Author:  Barry Minor, IBM AWS Graphics Systems (Austin)
*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>
#include <GL/gl.h>
#include "aux.h"
#include "viewperf.h"
#include "macros.h"
#include "bartfont2.h"

void (*func[10])();
void (*eventloop)(int, int, int);

char   rendermodetext[][16] = {
    "POINT",
    "VECTOR",
    "LINE",
    "POLYGON",
    "TFAN",
    "TMESH"
};

char   teststring[][32] = {
    "color/frame",
    "color/primitive",
    "color/vertex"
};

char   extention[][8] = {
    ".coor",
    ".elem",
    ".vnorm"
};

char   falsetrue[][8] = {
    "FALSE",
    "TRUE"
};

static unsigned char stipple[4*32] = {
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,

    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,

    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,
    0x00, 0xFF, 0xFF, 0x00,

    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
};

static float trans[4];
static float maxdim;
static float center[3];
static int vertsperframe;
static int tc;
static int np;
static float vert[MAXVERTS][3];
static float vnorm[MAXVERTS][3];
static float texture[MAXVERTS][2];
static float vcolor[MAXVERTS][3];
static struct plygon ply[MAXVERTS];
static struct mesh msh[MAXVERTS];

void Swap32(void *ptr, long length)
{
  register GLuint tmp;		/* GLuint should be typedef'ed to 32 bit
				   unsigned int */
  GLuint i, *array = (GLuint *) ptr;

  for (i = 0; i < length; i++) {
    tmp = array[i];
    tmp &= 0xffffffff;
    tmp = ((tmp >> 24) | (tmp << 24) |
	   ((tmp >> 8) & 0xff00) | ((tmp << 8) & 0xff0000));
    array[i] = tmp;
  }

}
char *error2str(GLenum err)
{
switch(err)
	{
	case GL_INVALID_ENUM: return("GL_INVALID_ENUM");
	case GL_INVALID_VALUE: return("GL_INVALID_VALUE");
	case GL_INVALID_OPERATION: return("GL_INVALID_OPERATION");
	case GL_STACK_OVERFLOW: return("GL_STACK_OVERFLOW");
	case GL_STACK_UNDERFLOW: return("GL_STACK_UNDERFLOW");
	case GL_OUT_OF_MEMORY: return("GL_OUT_OF_MEMORY");
	}
}

hsv_to_rgb(float h, float s, float v, float *r, float *g, float *b)
{
    int         i;
    float       f, p, q, t;

    if (s == 0) {
        *r = v;
        *g = v;
        *b = v;
    } else {
        h = fmod(h, 1.);
        h *= 6.;
        i = (int) floor(h);
        f = h - i;
        p = v*(1 - s);
        q = v*(1 - (s*f));
        t = v*(1 - (s*(1 - f)));
        switch (i) {
        case 0 :
            *r = v;
            *g = t;
            *b = p;
            break;
        case 1 :
            *r = q;
            *g = v;
            *b = p;
            break;
        case 2 :
            *r = p;
            *g = v;
            *b = t;
            break;
        case 3 :
            *r = p;
            *g = q;
            *b = v;
            break;
        case 4 :
            *r = t;
            *g = p;
            *b = v;
            break;
        case 5 :
            *r = v;
            *g = p;
            *b = q;
            break;
        }
    }
}

int param_poly(int numverts)
{
    int i;
    double phi,theta;
    double basex,basey,basemag;
    double uu,ww;

    for(i=1;i<=numverts;++i)
    {
        /* calculate location in texture map */
        phi=acos(((double) vnorm[i][2])); 
	basemag = sqrt(vnorm[i][0]*vnorm[i][0] + vnorm[i][1]*vnorm[i][1]);
	if (basemag == 0.0) {
	  basex = 0.0;
	  basey = 0.0;
	}
	else {
	  basex = vnorm[i][0]/basemag;
	  basey = vnorm[i][1]/basemag;
	}
	theta=asin(basey);
	if(basex < 0.0) theta=PI-theta;
	if(theta < 0.0) theta+=2.0*PI;
        uu=theta/(2.0*PI);
        ww=phi/(PI);
        texture[i][0] = uu;
        texture[i][1] = ww;
    }
}

int param_mesh()
{
    int i,k;
    float phi,theta;
    double basex,basey,basemag;
    float uu,ww;

    for(i=0;i<np;++i)
    {
        for(k=0;k<msh[i].numverts;++k)
        {
             /* calculate location in texture map */
            phi=acos(msh[i].norms[k].z);
            basemag = sqrt(msh[i].norms[k].x*msh[i].norms[k].x + msh[i].norms[k].y*msh[i].norms[k].y);
	    if (basemag == 0.0) {
	      basex = 0.0;
	      basey = 0.0;
	    }
	    else {
	      basex = msh[i].norms[k].x/basemag;
	      basey = msh[i].norms[k].y/basemag;
	    }
            theta=asin(basey);
            if(basex < 0.0) theta=PI-theta;
            if(theta < 0.0) theta+=2.0*PI;
            uu=theta/(2.0*PI);
            ww=phi/(PI);
            msh[i].texture[k].x = uu;
            msh[i].texture[k].y = ww;
        }
    }
}

/************************************************************************/
/************************************************************************/
/****                                                                ****/
/****  Begin Custom Rendering Routines                               ****/
/****                                                                ****/
/************************************************************************/
/************************************************************************/

#include "poly_set.h"
EXPAND(MULTI_,GL_POINTS,rmpoint);
EXPAND(SINGLE_,GL_LINE_STRIP,rmvector);
EXPAND(MULTI_,GL_LINE_LOOP,rmline);
EXPAND(MULTI_,GL_POLYGON,rmpoly);
EXPAND(MULTI_,GL_TRIANGLE_FAN,rmtfan);
#include "mesh_set.h"
EXPAND(MESH_,GL_POINTS,rmpmesh);
EXPAND(SINGLE_MESH_,GL_LINE_STRIP,rmvmesh);
EXPAND(MESH_,GL_LINE_STRIP,rmlmesh);
EXPAND(MESH_,GL_TRIANGLE_STRIP,rmtmesh);

/************************************************************************/
/************************************************************************/
/****                                                                ****/
/****  End Custom Rendering Routines                                 ****/
/****                                                                ****/
/************************************************************************/
/************************************************************************/

void meshinput(char *objnameptr)
{
    union {
	int testWord;
	char testByte[4];
    } endianTest;
    GLenum swapFlag;
    int i,k;
    FILE *f;
    char filename[50], *filenameptr;
    int tempspace[100];
    float maxmag,minmag,vertmag;
    float tmp;
    GLint cvcount;
    int nRead;
    float maxmin[3][2];

    endianTest.testWord = 1;
    if (endianTest.testByte[0] == 1) {
	swapFlag = GL_TRUE;
    } else {
	swapFlag = GL_FALSE;
    }
    np = 0;

    filenameptr = filename;

    for(i=0;i<3;++i)
    {
        maxmin[i][0] = SMALL;
        maxmin[i][1] = BIG;
    }

    filenameptr = strcpy(filenameptr, objnameptr);

    if((f = fopen(filename,"r")) == 0)
    {
        FATAL_ERROR("Couldn't open (mesh) input file\n");
    }

    while ((nRead = fread(&cvcount, sizeof(GLint), 1, f)) > 0) {
	if (swapFlag) Swap32((void *)&cvcount, 1);
        msh[np].numverts = cvcount;
        msh[np].verts = (struct vector *) malloc(cvcount*sizeof(struct vector));
        msh[np].norms = (struct vector *) malloc(cvcount*sizeof(struct vector));
        msh[np].texture = (struct vector *) malloc(cvcount*sizeof(struct vector));
        msh[np].vcolor = (struct colorvector *) malloc(cvcount*sizeof(struct colorvector));
        nRead = fread(&msh[np].verts[0].x, cvcount*sizeof(float)*3, 1, f);
	if (nRead == 0) {
	  FATAL_ERROR("unexpected end of input\n");
	}
	if (nRead < 0) {
	  FATAL_ERROR("read error\n");
	}
        nRead = fread(&msh[np].norms[0].x, cvcount*sizeof(float)*3, 1, f);
	if (nRead == 0) {
	  FATAL_ERROR("unexpected end of input\n");
	}
	if (nRead < 0) {
	  FATAL_ERROR("read error\n");
	}
	if (swapFlag) {
	  Swap32((void *)&msh[np].verts[0].x, cvcount*3);
	  Swap32((void *)&msh[np].norms[0].x, cvcount*3);
	}
        for(i=0;i<cvcount;++i)
        {
            if(msh[np].verts[i].x > maxmin[0][0])
                maxmin[0][0] = msh[np].verts[i].x;
            else if(msh[np].verts[i].x < maxmin[0][1])
                maxmin[0][1] = msh[np].verts[i].x;

            if(msh[np].verts[i].y > maxmin[1][0])
                maxmin[1][0] = msh[np].verts[i].y;
            else if( msh[np].verts[i].y < maxmin[1][1])
                maxmin[1][1] = msh[np].verts[i].y;

            if(msh[np].verts[i].z > maxmin[2][0])
                maxmin[2][0] = msh[np].verts[i].z;
            else if(msh[np].verts[i].z < maxmin[2][1])
                maxmin[2][1] = msh[np].verts[i].z;
        }
        ++np;
    }
    if (nRead < 0) {
      FATAL_ERROR("read error\n");
    }
    fclose(f);

    trans[3] = SMALL;
    for(i=0;i<3;i++)
    {
        tmp = maxmin[i][0] - maxmin[i][1];
        if(tmp > trans[3]) trans[3] = tmp;
    }

    center[0] = (maxmin[0][0]-maxmin[0][1]);
    center[1] = (maxmin[1][0]-maxmin[1][1]);
    center[2] = (maxmin[2][0]-maxmin[2][1]);

    maxdim=sqrt(center[0]*center[0]+center[1]*center[1]+center[2]*center[2])/2.0;

    center[0] /= 2.0;   /* don't move look up - need half dim*/
    center[1] /= 2.0;
    center[2] /= 2.0;
    trans[0] = maxmin[0][1] + center[0];
    trans[1] = maxmin[1][1] + center[1];
    trans[2] = maxmin[2][1] + center[2];

    maxmag = SMALL;
    minmag = BIG;
    for(i=0;i<np;++i)
    {
        for(k=0;k<msh[i].numverts;++k)
        {
            vertmag=sqrt((msh[i].verts[k].x-trans[0])*(msh[i].verts[k].x-trans[0])+
                (msh[i].verts[k].y-trans[1])*(msh[i].verts[k].y-trans[1])+
                (msh[i].verts[k].z-trans[2])*(msh[i].verts[k].z-trans[2]));
            if(vertmag > maxmag) maxmag = vertmag;
            if(vertmag < minmag) minmag = vertmag;
        }
    }
    for(i=0;i<np;++i)
    {
        for(k=0;k<msh[i].numverts;++k)
        {
            float r,g,b,d;

            vertmag=sqrt((msh[i].verts[k].x-trans[0])*(msh[i].verts[k].x-trans[0])+
                (msh[i].verts[k].y-trans[1])*(msh[i].verts[k].y-trans[1])+
                (msh[i].verts[k].z-trans[2])*(msh[i].verts[k].z-trans[2]));
            d = maxmag-vertmag;
	    if (d == 0.0)
	      vertmag = 1.0;
	    else
            vertmag = (maxmag-minmag)/d;
            hsv_to_rgb(vertmag, 1.0, 1.0, &r, &g, &b);
            msh[i].vcolor[k].r = r;
            msh[i].vcolor[k].g = g;
            msh[i].vcolor[k].b = b;
        }
    }
    param_mesh();
    vertsperframe = 0;
    for(i=0;i<np;++i)
    {
        vertsperframe += msh[i].numverts;
    }
}

void polygoninput(char *objnameptr)
{
    int i;
    FILE *f;
    int numverts;
    char filename[50], *filenameptr;
    int tempspace[100];
    float maxmag,minmag,vertmag;
    float tmp;
    char name[1024];
    int nRead;
    int index;

    float maxmin[3][2];

    np = 0;

    filenameptr = filename;

    for(i=0;i<3;++i)
    {
        maxmin[i][0] = SMALL;
        maxmin[i][1] = BIG;
    }

    filenameptr = strcpy(filenameptr, objnameptr);
    filenameptr = strcat(filenameptr, extention[0]);

    if((f = fopen(filename,"r")) == 0)
    {
        FATAL_ERROR("Couldn't open (coor) input file\n");
    }

    numverts = 0;
    while((nRead = fscanf(f,"%d",&index)) == 1) {
        nRead = fscanf(f,",%f,%f,%f",&vert[index][0], &vert[index][1], &vert[index][2]);
	if (nRead == EOF) {
	  FATAL_ERROR("unexpected end of input\n");
	}
	if (nRead != 3) {
	  FATAL_ERROR("read error\n");
	}

        if(vert[index][0] > maxmin[0][0])
            maxmin[0][0] = vert[index][0];
        else if(vert[index][0] < maxmin[0][1])
            maxmin[0][1] = vert[index][0];

        if(vert[index][1] > maxmin[1][0])
            maxmin[1][0] = vert[index][1];
        else if(vert[index][1] < maxmin[1][1])
            maxmin[1][1] = vert[index][1];

        if(vert[index][2] > maxmin[2][0])
            maxmin[2][0] = vert[index][2];
        else if(vert[index][2] < maxmin[2][1])
            maxmin[2][1] = vert[index][2];

        numverts++;
    }
    if (nRead != EOF) {
      FATAL_ERROR("read error\n");
    }

    fclose(f);

    filenameptr = strcpy(filenameptr, objnameptr);
    filenameptr = strcat(filenameptr, extention[1]);

    if((f = fopen(filename,"r")) == 0)
    {
        FATAL_ERROR("Couldn't open (elem) input file\n");
    }

    while((nRead = fscanf(f,"%s ", name)) == 1) {

        i=0;
        while((nRead = fscanf(f,"%d ", &tempspace[i])) == 1) i++;
	
	if (i > 0) {
	  ply[np].numverts = i;
	  ply[np].index = (int *) malloc(i*sizeof(int));
	  for(i = 0; i < ply[np].numverts; i++)
            ply[np].index[i] = (int) tempspace[i];

        ++np;
      }
    }
    if (nRead != EOF) {
      FATAL_ERROR("read error\n");
    }

    fclose(f);

    filenameptr = strcpy(filenameptr, objnameptr);
    filenameptr = strcat(filenameptr, extention[2]);

    if((f = fopen(filename,"r")) == 0)
    {
        FATAL_ERROR("Couldn't open (vnorm) input file\n");
    }
    for(i=1;i<=numverts;++i)
        nRead = fscanf(f,"%f %f %f\n",&vnorm[i][0],&vnorm[i][1],&vnorm[i][2]);
	if (nRead != 3) {
	  FATAL_ERROR("read error\n");
	}
    fclose(f);

    trans[3] = SMALL;
    for(i=0;i<3;i++)
    {
        tmp = maxmin[i][0] - maxmin[i][1];
        if(tmp > trans[3]) trans[3] = tmp;
    }

    center[0] = (maxmin[0][0]-maxmin[0][1]);
    center[1] = (maxmin[1][0]-maxmin[1][1]);
    center[2] = (maxmin[2][0]-maxmin[2][1]);

    maxdim=sqrt(center[0]*center[0]+center[1]*center[1]+center[2]*center[2])/2.0;

    center[0] /= 2.0;   /* don't move look up - need half dim*/
    center[1] /= 2.0;
    center[2] /= 2.0;
    trans[0] = maxmin[0][1] + center[0];
    trans[1] = maxmin[1][1] + center[1];
    trans[2] = maxmin[2][1] + center[2];

    maxmag = SMALL;
    minmag = BIG;
    for(i=1;i<=numverts;++i)
    {
        vertmag=sqrt((vert[i][0]-trans[0])*(vert[i][0]-trans[0])+
            (vert[i][1]-trans[1])*(vert[i][1]-trans[1])+
            (vert[i][2]-trans[2])*(vert[i][2]-trans[2]));
        if(vertmag > maxmag) maxmag = vertmag;
        if(vertmag < minmag) minmag = vertmag;
    }
    for(i=1;i<=numverts;++i)
    {
        float r,g,b,d;

        vertmag=sqrt((vert[i][0]-trans[0])*(vert[i][0]-trans[0])+
            (vert[i][1]-trans[1])*(vert[i][1]-trans[1])+
            (vert[i][2]-trans[2])*(vert[i][2]-trans[2]));
	d = maxmag-vertmag;
	if (d == 0.0)
	  vertmag = 1.0;
	else
	  vertmag = (maxmag-minmag)/d;
        hsv_to_rgb(vertmag, 1.0, 1.0, &r, &g, &b);
        vcolor[i][0] = r;
        vcolor[i][1] = g;
        vcolor[i][2] = b;
    }
    param_poly(numverts);
    vertsperframe = 0;
    for(i=0;i<np;++i)
    {
        vertsperframe += ply[i].numverts;
    }
}

void eventloopdisplaylist(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}


void eventloopdisplaylistzbuffer(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopdisplaylistclip(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopdisplaylistzbufferclip(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopimmedzbuffer(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmedzbufferclip(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmedclip(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmed(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopdisplaylistdb(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopdisplaylistzbufferdb(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopdisplaylistclipdb(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif
        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopdisplaylistzbufferclipdb(int st, int et, int numframes)
{
    GLenum err;
    int list;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    list = glGenLists(tc);
    for(ii=st;ii<et;++ii)
    {
        glNewList((list+ii), GL_COMPILE);
        (*func[ii])();
        glEndList();
    }
    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            glCallList(list+ii);
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
    glDeleteLists(list, tc);
}

void eventloopimmedzbufferdb(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmedzbufferclipdb(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmedclipdb(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(-center[0], -center[1], -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void eventloopimmeddb(int st, int et, int numframes)
{
    GLenum err;
    int ii,kk;
    float iang=0.0, jang=0.0, kang=0.0;

    for(ii=st;ii<et;++ii)
    {
#ifndef DEMO
        iang = 0.0;
        jang = 0.0;
        kang = 0.0;
#endif

        startclock();
        for(kk=0;kk<numframes;++kk)
        {
            glPushMatrix();
            glTranslatef(0.0, 0.0, -trans[3]*1.45);
            glRotatef(jang, 0.0, 1.0, 0.0);
            glRotatef(iang, 1.0, 0.0, 0.0);
            glRotatef(kang, 0.0, 0.0, 1.0);
            glTranslatef(-trans[0], -trans[1]*0.98, -trans[2]);

            glClear(GL_COLOR_BUFFER_BIT);

            (*func[ii])();
            glFlush();

            auxSwapBuffers();
            glPopMatrix();
            iang -= 1.0;
            jang += 5.0;
            kang += 0.5;

        }
        stopclock (numframes, teststring[ii]);
        if (err=glGetError() )
            fprintf(stderr, "<<WARNING>>: glError %s. The above results may be invalid\n", error2str(err));
    }
}

void def_light_calc(int nl, float loclight, float locview)
{
    float props[4];

    props[0] = 0.4;
    props[1] = 0.4;
    props[2] = 0.4;
    props[3] = 1.0;
    glMaterialfv(GL_FRONT, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glMaterialfv(GL_FRONT, GL_SPECULAR, props);
    props[0] = 0.7;
    props[1] = 0.7;
    props[2] = 0.7;
    props[3] = 1.0;
    glMaterialfv(GL_FRONT, GL_DIFFUSE, props);
    props[0] = 128;
    glMaterialfv(GL_FRONT, GL_SHININESS, props);

    props[0] = 1.000;
    props[1] = 1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT0, GL_POSITION, props);
    props[0] = 1.0/(float) nl;
    props[1] = 1.0/(float) nl;
    props[2] = 1.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT0, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT0, GL_SPECULAR, props);

    props[0] = 1.000;
    props[1] = 0.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT1, GL_POSITION, props);
    props[0] = 0.0/(float) nl;
    props[1] = 1.0/(float) nl;
    props[2] = 1.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT1, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT1, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT1, GL_SPECULAR, props);

    props[0] = 1.000;
    props[1] = -1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT2, GL_POSITION, props);
    props[0] = 1.0/(float) nl;
    props[1] = 0.0/(float) nl;
    props[2] = 1.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT2, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT2, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT2, GL_SPECULAR, props);

    props[0] = 0.000;
    props[1] = -1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT3, GL_POSITION, props);
    props[0] = 1.0/(float) nl;
    props[1] = 1.0/(float) nl;
    props[2] = 0.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT3, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT3, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT3, GL_SPECULAR, props);

    props[0] = -1.000;
    props[1] = -1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT4, GL_POSITION, props);
    props[0] = 1.0/(float) nl;
    props[1] = 0.0/(float) nl;
    props[2] = 0.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT4, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT4, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT4, GL_SPECULAR, props);

    props[0] = -1.000;
    props[1] = 0.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT5, GL_POSITION, props);
    props[0] = 0.0/(float) nl;
    props[1] = 1.0/(float) nl;
    props[2] = 0.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT5, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT5, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT5, GL_SPECULAR, props);

    props[0] = -1.000;
    props[1] = 1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT6, GL_POSITION, props);
    props[0] = 0.0/(float) nl;
    props[1] = 0.0/(float) nl;
    props[2] = 1.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT6, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT6, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT6, GL_SPECULAR, props);

    props[0] = 0.000;
    props[1] = 1.000;
    props[2] = 1.000;
    props[3] = loclight;
    glLightfv(GL_LIGHT7, GL_POSITION, props);
    props[0] = 1.0/(float) nl;
    props[1] = 1.0/(float) nl;
    props[2] = 1.0/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT7, GL_DIFFUSE, props);
    props[0] = 0.1/(float) nl;
    props[1] = 0.1/(float) nl;
    props[2] = 0.1/(float) nl;
    props[3] = 1.0;
    glLightfv(GL_LIGHT7, GL_AMBIENT, props);
    props[0] = 1.0;
    props[1] = 1.0;
    props[2] = 1.0;
    props[3] = 1.0;
    glLightfv(GL_LIGHT7, GL_SPECULAR, props);


    props[0] = 0.5;
    props[1] = 0.5;
    props[2] = 0.5;
    props[3] = 1.0;
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, props);
    glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, locview);
}

DrawString(char *str, float x, float y)
{
    short nseg, nstroke;
    short *cptr;
    int  i, j;

    glPushMatrix();
    glTranslatef (x, y, 0.0);
    glScalef (10.0, 10.0, 10.0);


    while (*str) {
        cptr = &(chrtbl[*str][0]);
        nseg = *(cptr++);
        for (i=0; i<nseg; i++) {
            nstroke = *(cptr++);
            glBegin(GL_LINE_STRIP);
            for (j=0; j<nstroke; j++) {
                glVertex2sv(cptr);
                cptr += 2;
            }
            glEnd();
        }
        glTranslatef (6.0, 0.0, 0.0);
        str++;
     }
    glPopMatrix();
}

main(int argc,char *argv[])
{
    GLenum type;
    int i,j,k;

    FILE *f,*texfile;

    int numframes=60;
    int testnum=0;
    int st,et;
    int numlights=0;
    int xsize=960,ysize=768;

    int inputmode=NOINPUT;
    int facetnormalmode=FALSE;
    int linesmoothmode=FALSE;
    int backfacemode=FALSE;
    int zbuffermode=FALSE;
    int fogmode=FALSE;
    int clipmode=FALSE;
    int shademodelmode=GL_SMOOTH;
    int polystipplemode = FALSE;
    int linestipplemode = FALSE;
    int doublebuffermode = TRUE;
    int directrendermode = FALSE;
    int texturemode = FALSE;
    int texenvmode = GL_DECAL;
    int rendermode = LINEmode;
    int prpmode = FALSE;
    int displaylistmode = FALSE;
    float linewidthmode = 1.0;
    float localviewmode = 0.0;
    float locallightmode = 0.0;

    char filename[50], *filenameptr;
    char objname[50], *objnameptr;

    char header[20];
    char *depth;
    long pack;
    int row,col,index;
    int xpix,ypix;
    int numpix;
    int numcomp = 3;
    float minfilter = GL_NEAREST;
    float magfilter = GL_NEAREST;
    float wrap = GL_CLAMP;
    unsigned char *Image;
    unsigned char *TextureImage;

    filenameptr = filename;
    objnameptr = objname;

    /* Parse command-line args. */
    while (--argc)
    {
        ++argv;

        /*** Options accepting arguments. ***/
        if (strcmp ("-backface", argv[0]) == 0 ||
            strcmp ("-bf", argv[0]) == 0)
        {
            backfacemode = TRUE;
        }
        else if (strcmp ("-rendermode", argv[0]) == 0 ||
            strcmp ("-rm", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -rendermode flag requires an argument.\n");
            if (strcmp ("VECTOR", argv[1]) == 0)
                rendermode = VECTORmode;
            else if (strcmp ("LINE", argv[1]) == 0)
                rendermode = LINEmode;
            else if (strcmp ("POLYGON", argv[1]) == 0)
                rendermode = POLYGONmode;
            else if (strcmp ("TFAN", argv[1]) == 0)
                rendermode = TFANmode;
            else if (strcmp ("POINT", argv[1]) == 0)
                rendermode = POINTmode;
            else if (strcmp ("TMESH", argv[1]) == 0)
                rendermode = TMESHmode;
            ARG_INC;
        }
        else if (strcmp ("-linesmooth", argv[0]) == 0 ||
            strcmp ("-ls", argv[0]) == 0)
        {
            linesmoothmode = TRUE;
        }
        else if (strcmp ("-facetnormal", argv[0]) == 0 ||
            strcmp ("-fn", argv[0]) == 0)
        {
            facetnormalmode = TRUE;
        }
        else if (strcmp ("-displaylist", argv[0]) == 0 ||
            strcmp ("-dl", argv[0]) == 0)
        {
            displaylistmode = TRUE;
        }
        else if (strcmp ("-pushrotpop", argv[0]) == 0 ||
            strcmp ("-prp", argv[0]) == 0)
        {
            prpmode = TRUE;
        }
        else if (strcmp ("-flat", argv[0]) == 0 ||
            strcmp ("-f", argv[0]) == 0)
        {
            shademodelmode = GL_FLAT;
        }
        else if (strcmp ("-directrender", argv[0]) == 0 ||
            strcmp ("-dr", argv[0]) == 0)
        {
            directrendermode = TRUE;
        }
        else if (strcmp ("-singlebuffer", argv[0]) == 0 ||
            strcmp ("-sb", argv[0]) == 0)
        {
            doublebuffermode = FALSE;
        }
        else if (strcmp ("-zbuffer", argv[0]) == 0 ||
            strcmp ("-zb", argv[0]) == 0)
        {
            zbuffermode = TRUE;
        }
        else if (strcmp ("-linestipple", argv[0]) == 0 ||
            strcmp ("-lp", argv[0]) == 0)
        {
            linestipplemode = TRUE;
        }
        else if (strcmp ("-polystipple", argv[0]) == 0 ||
            strcmp ("-pp", argv[0]) == 0)
        {
            polystipplemode = TRUE;
        }
        else if (strcmp ("-fog", argv[0]) == 0 ||
            strcmp ("-fg", argv[0]) == 0)
        {
            fogmode = TRUE;
        }
        else if (strcmp ("-clip", argv[0]) == 0 ||
            strcmp ("-c", argv[0]) == 0)
        {
            clipmode = TRUE;
        }
        else if (strcmp ("-localview", argv[0]) == 0 ||
            strcmp ("-lv", argv[0]) == 0)
        {
            localviewmode = 1.0;
        }
        else if (strcmp ("-locallight", argv[0]) == 0 ||
            strcmp ("-ll", argv[0]) == 0)
        {
            locallightmode = 1.0;
        }
        else if (strcmp ("-testnum", argv[0]) == 0 ||
            strcmp ("-tn", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -testnum flag requires an argument.\n");
            testnum = atoi (argv[1]);
            if(testnum < 1 || testnum > 3)
                testnum = 3;
            ARG_INC;
        }
        else if (strcmp ("-texcomp", argv[0]) == 0 ||
            strcmp ("-tc", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -texcomp flag requires an argument.\n");
            numcomp = atoi (argv[1]);
            if(numcomp < 1 || numcomp > 4)
                numcomp = 3;
            ARG_INC;
        }
        else if (strcmp ("-texenvmode", argv[0]) == 0 ||
            strcmp ("-te", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -texenv flag requires an argument.\n");
            if (strcmp ("MODULATE", argv[1]) == 0)
                texenvmode = GL_MODULATE;
            else if (strcmp ("BLEND", argv[1]) == 0)
                texenvmode = GL_BLEND;
            else if (strcmp ("DECAL", argv[1]) == 0)
                texenvmode = GL_DECAL;
            ARG_INC;
        }
        else if (strcmp ("-texture", argv[0]) == 0 ||
            strcmp ("-tx", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -texture flag requires an argument.\n");
            texturemode = TRUE;
            if ((texfile = fopen(argv[1],"r") ) == 0)
            {
                FATAL_ERROR("Can't open texture file\n");
            }
            ARG_INC;
        }
        else if (strcmp ("-minfilter", argv[0]) == 0 ||
            strcmp ("-minf", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -minfilter flag requires an argument.\n");
            if (strcmp ("NEAREST", argv[1]) == 0)
                minfilter = GL_NEAREST;
            else if (strcmp ("LINEAR", argv[1]) == 0)
                minfilter = GL_LINEAR;
            else if (strcmp ("NEAREST_MIPMAP_NEAREST",argv[1]) == 0)
                minfilter = GL_NEAREST_MIPMAP_NEAREST;
            else if (strcmp ("LINEAR_MIPMAP_LINEAR", argv[1]) == 0)
                minfilter = GL_LINEAR_MIPMAP_LINEAR;
            else if (strcmp ("NEAREST_MIPMAP_LINEAR", argv[1]) == 0)
                minfilter = GL_NEAREST_MIPMAP_LINEAR;
            else if (strcmp ("LINEAR_MIPMAP_LINEAR", argv[1]) == 0)
                minfilter = GL_LINEAR_MIPMAP_LINEAR;
            ARG_INC;
        }
        else if (strcmp ("-magfilter", argv[0]) == 0 ||
            strcmp ("-magf", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -magfilter flag requires an argument.\n");
            if (strcmp ("NEAREST", argv[1]) == 0)
                magfilter = GL_NEAREST;
            else if (strcmp ("LINEAR", argv[1]) == 0)
                magfilter = GL_LINEAR;
            ARG_INC;
        }
        else if (strcmp ("-mesh", argv[0]) == 0 ||
            strcmp ("-mh", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -mesh flag requires an argument.\n");
            objnameptr = strcpy(objnameptr, argv[1]);
            inputmode = MESHINPUT;
            ARG_INC;
        }
        else if (strcmp ("-object", argv[0]) == 0 ||
            strcmp ("-ob", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -object flag requires an argument.\n");
            objnameptr = strcpy(objnameptr, argv[1]);
            inputmode = POLYGONINPUT;
            ARG_INC;
        }
        else if (strcmp ("-linewidth", argv[0]) == 0 ||
            strcmp ("-lw", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -linewidth flag requires an argument.\n");
            linewidthmode = atof (argv[1]);
            ARG_INC;
        }
        else if (strcmp ("-xwinsize", argv[0]) == 0 ||
            strcmp ("-xws", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -xwinsize flag requires an argument.\n");
            xsize = atoi (argv[1]);
            ARG_INC;
        }
        else if (strcmp ("-ywinsize", argv[0]) == 0 ||
            strcmp ("-yws", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -ywinsize flag requires an argument.\n");
            ysize = atoi (argv[1]);
            ARG_INC;
        }
        else if (strcmp ("-numframes", argv[0]) == 0 ||
            strcmp ("-nf", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -numframes flag requires an argument.\n");
            numframes = atoi (argv[1]);
            ARG_INC;
        }
        else if (strcmp ("-numlights", argv[0]) == 0 ||
            strcmp ("-nl", argv[0]) == 0)
        {
            if (argc == 1)
                FATAL_ERROR ("viewperf: the -numlights flag requires an argument.\n");
            numlights = atoi (argv[1]);
            ARG_INC;
        }
        else if (strcmp ("-help", argv[0]) == 0 ||
            strcmp ("-usage", argv[0]) == 0 ||
            strcmp ("-h", argv[0]) == 0 ||
            strcmp ("-u", argv[0]) == 0)
        {
            fprintf (stderr, "viewperf version 1.2:\n\
Program options:\n\
-object -ob    <file>  : Viewpoint object to be used in the tests\n\
-mesh -mh      <file>  : Mesh object to be used in the tests\n\
-rendermode -rm <mode> : POINT, VECTOR, LINE, POLYGON, TMESH, or TFAN\n\
                         - default LINE\n\
-directrender -dr      : Render direct - default indirect\n\
-displaylist -dl       : Render with display list mode\n\
-testnum -tn   <num>   : 1 = Color/Frame, 2 = Color/Primitive, 3 = Color/Vert\n\
-texture -tx   <file>  : MTV image for texturing\n\
-magfilter -magf <flt> : NEAREST, LINEAR - default NEAREST\n\
-minfilter -minf <flt> : NEAREST, LINEAR, NEAREST_MIPMAP_NEAREST,\n\
                         LINEAR_MIPMAP_NEAREST, NEAREST_MIPMAP_LINEAR,\n\
                         LINEAR_MIPMAP_LINEAR - default NEAREST\n\
-texenv -te <env>      : Texture environment, MODULATE, DECAL, BLEND\n\
                         - default DECAL\n\
-texcomp -tc <num>     : Texture components where <num> is 1,2,3, or 4\n\
                       : -default 3\n\
-linewidth -lw <width> : Linewidth for wire/vector tests - default 1.0\n\
-xwinsize -xws <side>  : Size of test windows X dimension - default 960\n\
-ywinsize -yws <side>  : Size of test windows Y dimension - default 768\n\
-numframes -nf <num>   : Number of frames to be render per test - default 60\n\
-numlights -nl <num>   : Turns on <num> lights - default 0\n\
-backface -bf          : Backface primitives in test run - default off\n\
-singlebuffer -sb      : Single buffer mode\n\
-fog -fg               : Turn on fog\n\
-linesmooth -ls        : Turn on line antialiasing\n\
-facetnormal -fn       : Use facet normals when lighting\n\
-linestipple -lp       : Turn on line stipple\n\
-polystipple -pp       : Turn on polygon stipple\n\
-flat -f               : Set shademodel to FLAT - default GOURAUD\n\
-zbuffer -zb           : Turn zbuffer on for tests - default off\n\
-clip -c               : Align object on 3D clip boundary\n\
-pushrotpop -prp       : pushmatrix rot popmatrix before each poly or line\n\
-localview -lv         : Define local viewer for lit tests\n\
-locallight -ll        : Define all lights as local for lit tests\n\
");
            exit(0);
        }
    }

    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/
    /*                                                                   */
    /*                      Open Window                                  */
    /*                                                                   */
    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/

    auxInitPosition(0,0,xsize,ysize);
    type = AUX_DEPTH;
    type |= AUX_RGB;
    type |= (doublebuffermode) ? AUX_DOUBLE : AUX_SINGLE;
    type |= (directrendermode) ? AUX_DIRECT : AUX_INDIRECT;
    auxInitDisplayMode(type);

    if (auxInitWindow("Viewperf") == GL_FALSE) {
        auxQuit();
    }

/*
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);
*/

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1280.0, 0.0, 1024.0, 1.0, -1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glShadeModel(GL_SMOOTH);

    glBegin(GL_QUADS);
    glColor3f(1.0, 0.9, 0.3);
    glVertex2f(0.0, 0.0);
    glColor3f(1.0, 0.9, 0.3);
    glVertex2f(1280.0, 0.0);
    glColor3f(0.6, 0.1, 0.9);
    glVertex2f(1280.0, 1024.0);
    glColor3f(0.6, 0.1, 0.9);
    glVertex2f(0.0, 1024.0);
    glEnd();

    glLineWidth(4.0);
    glColor3f(0.0, 0.0, 0.0);
    DrawString("     OpenGL     ", 172.0, 602.0);
    glLineWidth(3.0);
    glColor3f(1.0, 1.0, 0.0);
    DrawString("     OpenGL     ", 170.0, 600.0);

    glLineWidth(4.0);
    glColor3f(0.0, 0.0, 0.0);
    DrawString("    Viewperf    ", 172.0, 477.0);
    glLineWidth(3.0);
    glColor3f(0.0, 1.0, 0.0);
    DrawString("    Viewperf    ", 170.0, 475.0);

    glLineWidth(4.0);
    glColor3f(0.0, 0.0, 0.0);
    DrawString("Loading Data Set", 172.0, 352.0);
    glLineWidth(3.0);
    glColor3f(0.0, 0.7, 1.0);
    DrawString("Loading Data Set", 170.0, 350.0);

    if(doublebuffermode)
        auxSwapBuffers();
    sleep(1);

    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/
    /*                                                                   */
    /*                        Read Data                                  */
    /*                                                                   */
    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/

    switch(inputmode)
    {
    case NOINPUT:
        fprintf(stderr,"No input file specified\n");
        exit(0);
        break;

    case POLYGONINPUT:
        polygoninput(objnameptr);
        break;

    case MESHINPUT:
        meshinput(objnameptr);
        break;
    }

    fprintf (stdout, "viewperf: State Data\n\n\
-object\t\t\t%s - %d Vertices/Frame - %d Primitives\n\
-object\t\t\t\t - %f Vertices/Primitive\n\
-rendermode\t\t%s\n\
-numframes\t\t%d\n\
-xwinsize\t\t%d\n\
-ywinsize\t\t%d\n\
-directrender\t\t%s\n\
-displaylist\t\t%s\n\
-pushrotpop\t\t%s\n\
-clip\t\t\t%s\n\
-doublebuffer\t\t%s\n\
-backface\t\t%s\n\
-zbuffer\t\t%s\n\
-flat\t\t\t%s\n\
-fog\t\t\t%s\n\
-linestipple\t\t%s\n\
-polystipple\t\t%s\n\
-linesmooth\t\t%s\n\
-linewidth\t\t%f\n\n\
-numlights\t\t%d\n\
-facetnormal\t\t%s\n\
-localview\t\t%s\n\
-locallight\t\t%s\n\n\
",
        objnameptr,
        vertsperframe,
        np,
        ((float) vertsperframe)/((float) np),
        rendermodetext[rendermode],
        numframes,
        xsize,
        ysize,
        falsetrue[directrendermode],
        falsetrue[displaylistmode],
        falsetrue[prpmode],
        falsetrue[clipmode],
        falsetrue[doublebuffermode],
        falsetrue[backfacemode],
        falsetrue[zbuffermode],
        falsetrue[shademodelmode==GL_FLAT],
        falsetrue[fogmode],
        falsetrue[linestipplemode],
        falsetrue[polystipplemode],
        falsetrue[linesmoothmode],
        linewidthmode,
        numlights, 
        falsetrue[facetnormalmode],
        falsetrue[(int)localviewmode],
        falsetrue[(int)locallightmode]);

    if(inputmode == MESHINPUT)
    {
        switch(rendermode)
        {
        case TMESHmode:
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtmeshfnprptexture;
                                func[tc++] = rmtmeshfncolorprptexture;
                                func[tc++] = rmtmeshfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmtmeshfnprp;
                                func[tc++] = rmtmeshfncolorprp;
                                func[tc++] = rmtmeshfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtmeshfntexture;
                                func[tc++] = rmtmeshfncolortexture;
                                func[tc++] = rmtmeshfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmtmeshfn;
                                func[tc++] = rmtmeshfncolor;
                                func[tc++] = rmtmeshfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtmeshvnprptexture;
                                func[tc++] = rmtmeshvncolorprptexture;
                                func[tc++] = rmtmeshvnvertcolorprptexture;
                                func[tc++] = rmtmeshvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmtmeshvnprp;
                                func[tc++] = rmtmeshvncolorprp;
                                func[tc++] = rmtmeshvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtmeshvntexture;
                                func[tc++] = rmtmeshvncolortexture;
                                func[tc++] = rmtmeshvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmtmeshvn;
                                func[tc++] = rmtmeshvncolor;
                                func[tc++] = rmtmeshvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmtmeshprptexture;
                            func[tc++] = rmtmeshcolorprptexture;
                            func[tc++] = rmtmeshvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmtmeshprp;
                            func[tc++] = rmtmeshcolorprp;
                            func[tc++] = rmtmeshvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmtmeshtexture;
                            func[tc++] = rmtmeshcolortexture;
                            func[tc++] = rmtmeshvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmtmesh;
                            func[tc++] = rmtmeshcolor;
                            func[tc++] = rmtmeshvertcolor;
                        }
                    }

                }
            }
            break;
        case VECTORmode:
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvmeshfnprptexture;
                                func[tc++] = rmvmeshfncolorprptexture;
                                func[tc++] = rmvmeshfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmvmeshfnprp;
                                func[tc++] = rmvmeshfncolorprp;
                                func[tc++] = rmvmeshfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvmeshfntexture;
                                func[tc++] = rmvmeshfncolortexture;
                                func[tc++] = rmvmeshfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmvmeshfn;
                                func[tc++] = rmvmeshfncolor;
                                func[tc++] = rmvmeshfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvmeshvnprptexture;
                                func[tc++] = rmvmeshvncolorprptexture;
                                func[tc++] = rmvmeshvnvertcolorprptexture;
                                func[tc++] = rmvmeshvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmvmeshvnprp;
                                func[tc++] = rmvmeshvncolorprp;
                                func[tc++] = rmvmeshvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvmeshvntexture;
                                func[tc++] = rmvmeshvncolortexture;
                                func[tc++] = rmvmeshvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmvmeshvn;
                                func[tc++] = rmvmeshvncolor;
                                func[tc++] = rmvmeshvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmvmeshprptexture;
                            func[tc++] = rmvmeshcolorprptexture;
                            func[tc++] = rmvmeshvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmvmeshprp;
                            func[tc++] = rmvmeshcolorprp;
                            func[tc++] = rmvmeshvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmvmeshtexture;
                            func[tc++] = rmvmeshcolortexture;
                            func[tc++] = rmvmeshvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmvmesh;
                            func[tc++] = rmvmeshcolor;
                            func[tc++] = rmvmeshvertcolor;
                        }
                    }

                }
            }
            break;
        case LINEmode:
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlmeshfnprptexture;
                                func[tc++] = rmlmeshfncolorprptexture;
                                func[tc++] = rmlmeshfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmlmeshfnprp;
                                func[tc++] = rmlmeshfncolorprp;
                                func[tc++] = rmlmeshfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlmeshfntexture;
                                func[tc++] = rmlmeshfncolortexture;
                                func[tc++] = rmlmeshfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmlmeshfn;
                                func[tc++] = rmlmeshfncolor;
                                func[tc++] = rmlmeshfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlmeshvnprptexture;
                                func[tc++] = rmlmeshvncolorprptexture;
                                func[tc++] = rmlmeshvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmlmeshvnprp;
                                func[tc++] = rmlmeshvncolorprp;
                                func[tc++] = rmlmeshvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlmeshvntexture;
                                func[tc++] = rmlmeshvncolortexture;
                                func[tc++] = rmlmeshvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmlmeshvn;
                                func[tc++] = rmlmeshvncolor;
                                func[tc++] = rmlmeshvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmlmeshprptexture;
                            func[tc++] = rmlmeshcolorprptexture;
                            func[tc++] = rmlmeshvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmlmeshprp;
                            func[tc++] = rmlmeshcolorprp;
                            func[tc++] = rmlmeshvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmlmeshtexture;
                            func[tc++] = rmlmeshcolortexture;
                            func[tc++] = rmlmeshvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmlmesh;
                            func[tc++] = rmlmeshcolor;
                            func[tc++] = rmlmeshvertcolor;
                        }
                    }

                }
            }
            break;
        case POINTmode:
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpmeshfnprptexture;
                                func[tc++] = rmpmeshfncolorprptexture;
                                func[tc++] = rmpmeshfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpmeshfnprp;
                                func[tc++] = rmpmeshfncolorprp;
                                func[tc++] = rmpmeshfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpmeshfntexture;
                                func[tc++] = rmpmeshfncolortexture;
                                func[tc++] = rmpmeshfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpmeshfn;
                                func[tc++] = rmpmeshfncolor;
                                func[tc++] = rmpmeshfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpmeshvnprptexture;
                                func[tc++] = rmpmeshvncolorprptexture;
                                func[tc++] = rmpmeshvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpmeshvnprp;
                                func[tc++] = rmpmeshvncolorprp;
                                func[tc++] = rmpmeshvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpmeshvntexture;
                                func[tc++] = rmpmeshvncolortexture;
                                func[tc++] = rmpmeshvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpmeshvn;
                                func[tc++] = rmpmeshvncolor;
                                func[tc++] = rmpmeshvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpmeshprptexture;
                            func[tc++] = rmpmeshcolorprptexture;
                            func[tc++] = rmpmeshvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmpmeshprp;
                            func[tc++] = rmpmeshcolorprp;
                            func[tc++] = rmpmeshvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpmeshtexture;
                            func[tc++] = rmpmeshcolortexture;
                            func[tc++] = rmpmeshvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmpmesh;
                            func[tc++] = rmpmeshcolor;
                            func[tc++] = rmpmeshvertcolor;
                        }
                    }

                }
            }
            break;
        case POLYGONmode:
            fprintf(stderr,"Rendermode POLYGON not available for mesh objects\n");
            exit(0);
            break;
        case TFANmode:
            fprintf(stderr,"Rendermode TFAN not available for mesh objects\n");
            exit(0);
            break;
        }
    }
    else
    {
        switch(rendermode)
        {
        case POLYGONmode :
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpolyfnprptexture;
                                func[tc++] = rmpolyfncolorprptexture;
                                func[tc++] = rmpolyfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpolyfnprp;
                                func[tc++] = rmpolyfncolorprp;
                                func[tc++] = rmpolyfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpolyfntexture;
                                func[tc++] = rmpolyfncolortexture;
                                func[tc++] = rmpolyfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpolyfn;
                                func[tc++] = rmpolyfncolor;
                                func[tc++] = rmpolyfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpolyvnprptexture;
                                func[tc++] = rmpolyvncolorprptexture;
                                func[tc++] = rmpolyvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpolyvnprp;
                                func[tc++] = rmpolyvncolorprp;
                                func[tc++] = rmpolyvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpolyvntexture;
                                func[tc++] = rmpolyvncolortexture;
                                func[tc++] = rmpolyvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpolyvn;
                                func[tc++] = rmpolyvncolor;
                                func[tc++] = rmpolyvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpolyprptexture;
                            func[tc++] = rmpolycolorprptexture;
                            func[tc++] = rmpolyvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmpolyprp;
                            func[tc++] = rmpolycolorprp;
                            func[tc++] = rmpolyvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpolytexture;
                            func[tc++] = rmpolycolortexture;
                            func[tc++] = rmpolyvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmpoly;
                            func[tc++] = rmpolycolor;
                            func[tc++] = rmpolyvertcolor;
                        }
                    }

                }
            }
            break;
        case TFANmode :
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtfanfnprptexture;
                                func[tc++] = rmtfanfncolorprptexture;
                                func[tc++] = rmtfanfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmtfanfnprp;
                                func[tc++] = rmtfanfncolorprp;
                                func[tc++] = rmtfanfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtfanfntexture;
                                func[tc++] = rmtfanfncolortexture;
                                func[tc++] = rmtfanfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmtfanfn;
                                func[tc++] = rmtfanfncolor;
                                func[tc++] = rmtfanfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtfanvnprptexture;
                                func[tc++] = rmtfanvncolorprptexture;
                                func[tc++] = rmtfanvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmtfanvnprp;
                                func[tc++] = rmtfanvncolorprp;
                                func[tc++] = rmtfanvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmtfanvntexture;
                                func[tc++] = rmtfanvncolortexture;
                                func[tc++] = rmtfanvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmtfanvn;
                                func[tc++] = rmtfanvncolor;
                                func[tc++] = rmtfanvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmtfanprptexture;
                            func[tc++] = rmtfancolorprptexture;
                            func[tc++] = rmtfanvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmtfanprp;
                            func[tc++] = rmtfancolorprp;
                            func[tc++] = rmtfanvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmtfantexture;
                            func[tc++] = rmtfancolortexture;
                            func[tc++] = rmtfanvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmtfan;
                            func[tc++] = rmtfancolor;
                            func[tc++] = rmtfanvertcolor;
                        }
                    }

                }
            }
            break;
        case POINTmode :
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpointfnprptexture;
                                func[tc++] = rmpointfncolorprptexture;
                                func[tc++] = rmpointfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpointfnprp;
                                func[tc++] = rmpointfncolorprp;
                                func[tc++] = rmpointfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpointfntexture;
                                func[tc++] = rmpointfncolortexture;
                                func[tc++] = rmpointfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpointfn;
                                func[tc++] = rmpointfncolor;
                                func[tc++] = rmpointfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpointvnprptexture;
                                func[tc++] = rmpointvncolorprptexture;
                                func[tc++] = rmpointvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmpointvnprp;
                                func[tc++] = rmpointvncolorprp;
                                func[tc++] = rmpointvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmpointvntexture;
                                func[tc++] = rmpointvncolortexture;
                                func[tc++] = rmpointvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmpointvn;
                                func[tc++] = rmpointvncolor;
                                func[tc++] = rmpointvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpointprptexture;
                            func[tc++] = rmpointcolorprptexture;
                            func[tc++] = rmpointvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmpointprp;
                            func[tc++] = rmpointcolorprp;
                            func[tc++] = rmpointvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmpointtexture;
                            func[tc++] = rmpointcolortexture;
                            func[tc++] = rmpointvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmpoint;
                            func[tc++] = rmpointcolor;
                            func[tc++] = rmpointvertcolor;
                        }
                    }

                }
            }
            break;
        case VECTORmode :
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvectorfnprptexture;
                                func[tc++] = rmvectorfncolorprptexture;
                                func[tc++] = rmvectorfnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmvectorfnprp;
                                func[tc++] = rmvectorfncolorprp;
                                func[tc++] = rmvectorfnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvectorfntexture;
                                func[tc++] = rmvectorfncolortexture;
                                func[tc++] = rmvectorfnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmvectorfn;
                                func[tc++] = rmvectorfncolor;
                                func[tc++] = rmvectorfnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvectorvnprptexture;
                                func[tc++] = rmvectorvncolorprptexture;
                                func[tc++] = rmvectorvnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmvectorvnprp;
                                func[tc++] = rmvectorvncolorprp;
                                func[tc++] = rmvectorvnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmvectorvntexture;
                                func[tc++] = rmvectorvncolortexture;
                                func[tc++] = rmvectorvnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmvectorvn;
                                func[tc++] = rmvectorvncolor;
                                func[tc++] = rmvectorvnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmvectorprptexture;
                            func[tc++] = rmvectorcolorprptexture;
                            func[tc++] = rmvectorvertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmvectorprp;
                            func[tc++] = rmvectorcolorprp;
                            func[tc++] = rmvectorvertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmvectortexture;
                            func[tc++] = rmvectorcolortexture;
                            func[tc++] = rmvectorvertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmvector;
                            func[tc++] = rmvectorcolor;
                            func[tc++] = rmvectorvertcolor;
                        }
                    }

                }
            }
            break;
        case LINEmode :
            {
                if(numlights)
                {
                    if(facetnormalmode)
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlinefnprptexture;
                                func[tc++] = rmlinefncolorprptexture;
                                func[tc++] = rmlinefnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmlinefnprp;
                                func[tc++] = rmlinefncolorprp;
                                func[tc++] = rmlinefnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlinefntexture;
                                func[tc++] = rmlinefncolortexture;
                                func[tc++] = rmlinefnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmlinefn;
                                func[tc++] = rmlinefncolor;
                                func[tc++] = rmlinefnvertcolor;
                            }
                        }
                    }
                    else
                    {
                        if(prpmode)
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlinevnprptexture;
                                func[tc++] = rmlinevncolorprptexture;
                                func[tc++] = rmlinevnvertcolorprptexture;
                            }
                            else
                            {
                                func[tc++] = rmlinevnprp;
                                func[tc++] = rmlinevncolorprp;
                                func[tc++] = rmlinevnvertcolorprp;
                            }
                        }
                        else
                        {
                            if(texturemode)
                            {
                                func[tc++] = rmlinevntexture;
                                func[tc++] = rmlinevncolortexture;
                                func[tc++] = rmlinevnvertcolortexture;
                            }
                            else
                            {
                                func[tc++] = rmlinevn;
                                func[tc++] = rmlinevncolor;
                                func[tc++] = rmlinevnvertcolor;
                            }
                        }
                    }
                }
                else
                {
                    if(prpmode)
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmlineprptexture;
                            func[tc++] = rmlinecolorprptexture;
                            func[tc++] = rmlinevertcolorprptexture;
                        }
                        else
                        {
                            func[tc++] = rmlineprp;
                            func[tc++] = rmlinecolorprp;
                            func[tc++] = rmlinevertcolorprp;
                        }
                    }
                    else
                    {
                        if(texturemode)
                        {
                            func[tc++] = rmlinetexture;
                            func[tc++] = rmlinecolortexture;
                            func[tc++] = rmlinevertcolortexture;
                        }
                        else
                        {
                            func[tc++] = rmline;
                            func[tc++] = rmlinecolor;
                            func[tc++] = rmlinevertcolor;
                        }
                    }

                }
            }
            break;
        case TMESHmode:
            fprintf(stderr,"Rendermode TMESH not available for polygonal objects\n");
            exit(0);
            break;
        }
    }

    if(doublebuffermode)
    {
        if(displaylistmode)
        {
            if(zbuffermode)
            {
                if(clipmode)
                {
                    eventloop = eventloopdisplaylistzbufferclipdb;
                }
                else
                {
                    eventloop = eventloopdisplaylistzbufferdb;
                }
            }
            else if(clipmode)
            {
                eventloop = eventloopdisplaylistclipdb;
            }
            else
            {
                eventloop = eventloopdisplaylistdb;
            }
        }
        else
        {
            if(zbuffermode)
            {
                if(clipmode)
                {
                    eventloop = eventloopimmedzbufferclipdb;
                }
                else
                {
                    eventloop = eventloopimmedzbufferdb;
                }
            }
            else if(clipmode)
            {
                eventloop = eventloopimmedclipdb;
            }
            else
            {
                eventloop = eventloopimmeddb;
            }
        }
    }
    else
    {
        if(displaylistmode)
        {
            if(zbuffermode)
            {
                if(clipmode)
                {
                    eventloop = eventloopdisplaylistzbufferclip;
                }
                else
                {
                    eventloop = eventloopdisplaylistzbuffer;
                }
            }
            else if(clipmode)
            {
                eventloop = eventloopdisplaylistclip;
            }
            else
            {
                eventloop = eventloopdisplaylist;
            }
        }
        else
        {
            if(zbuffermode)
            {
                if(clipmode)
                {
                    eventloop = eventloopimmedzbufferclip;
                }
                else
                {
                    eventloop = eventloopimmedzbuffer;
                }
            }
            else if(clipmode)
            {
                eventloop = eventloopimmedclip;
            }
            else
            {
                eventloop = eventloopimmed;
            }
        }
    }


    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/
    /*                                                                   */
    /*                     Begin Rendering                               */
    /*                                                                   */
    /*********************************************************************/
    /*********************************************************************/
    /*********************************************************************/

    glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_FALSE);

    if(linestipplemode)
    {
        glLineStipple(1, 0xf0f0);
        glEnable(GL_LINE_STIPPLE);
    }

    if(polystipplemode)
    {
        glPolygonStipple(stipple);
        glEnable(GL_POLYGON_STIPPLE);
    }
    glShadeModel(shademodelmode);
    if(zbuffermode)
    {
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LESS);
        glDepthMask(GL_TRUE);
    }
    else
    {
        glDisable(GL_DEPTH_TEST);
        glDepthMask(GL_FALSE);
    }
    if(backfacemode)
    {
        glCullFace(GL_BACK);
        glEnable(GL_CULL_FACE);
    }
    glLineWidth(linewidthmode);
    if(linesmoothmode)
    {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
        glEnable(GL_LINE_SMOOTH);
    }
    if(fogmode)
    {
        glFogf(GL_FOG_START, trans[3]*1.45-maxdim);
        glFogf(GL_FOG_END, trans[3]*1.45+maxdim);
        glFogf(GL_FOG_MODE, GL_LINEAR);
        glEnable(GL_FOG);
    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, 1.250, trans[3]*1.45-maxdim, trans[3]*1.45+maxdim);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();


    if(numlights)
    {
        def_light_calc(numlights, locallightmode, localviewmode);
        glEnable(GL_LIGHTING);
        for(i=0;i<numlights;++i)
        {
            glEnable(GL_LIGHT0+i);
        }
    }

    if(texturemode)
    {
        k=0;
        do
        {
            fread(&header[k], 1, 1, texfile);
            ++k;
        } while ((header[k-1] != 0) && (header[k-1] != 10));

        for(i=0;i<k;++i)
            if(header[i]==' ')
            {
                header[i]=0;
                depth = &header[i+1];
                break;
            }

        xpix = atoi(header);
        ypix = atoi(depth);

        numpix = xpix*ypix;
        Image = (unsigned char *) malloc ( numpix*3 );
        TextureImage = (unsigned char *) malloc(numpix*numcomp);

        fread(Image, 3, numpix, texfile);
        fclose(f);

        switch(numcomp)
        {
        case 1 :
            i=0;
            for(j=0; j<numpix*3; j+=3)
            {
                TextureImage[i++] = Image[j+1];
            }
            break;
        case 2 :
            i=0;
            for(j=0; j<numpix*3; j+=3)
            {
                TextureImage[i++] = Image[j];
                TextureImage[i++] = 0x77;
            }
            break;
        case 3 :
            i=0;
            for(j=0; j<numpix*3; j+=3)
            {
                TextureImage[i++] = Image[j];
                TextureImage[i++] = Image[j+1];
                TextureImage[i++] = Image[j+2];
            }
            break;
        case 4 :
            i=0;
            for(j=0; j<numpix*3; j+=3)
            {
                TextureImage[i++] = Image[j];
                TextureImage[i++] = Image[j+1];
                TextureImage[i++] = Image[j+2];
                TextureImage[i++] = 0x77;
            }
            break;
        }
        free(Image);


        switch(numcomp)
        {
        case 1:
            if(gluBuild2DMipmaps(GL_TEXTURE_2D, numcomp, xpix, 
                ypix, GL_LUMINANCE, GL_UNSIGNED_BYTE, TextureImage)!=0)
            {
                fprintf(stderr,"Mipmaps didn't build\n");
                exit(0);
            }
            break;
        case 2:
            if(gluBuild2DMipmaps(GL_TEXTURE_2D, numcomp, xpix, 
                ypix, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, TextureImage)!=0)
            {
                fprintf(stderr,"Mipmaps didn't build\n");
                exit(0);
            }
            break;
        case 3:
            if(gluBuild2DMipmaps(GL_TEXTURE_2D, numcomp, xpix, 
                ypix, GL_RGB, GL_UNSIGNED_BYTE, TextureImage)!=0)
            {
                fprintf(stderr,"Mipmaps didn't build\n");
                exit(0);
            }
            break;
        case 4:
            if(gluBuild2DMipmaps(GL_TEXTURE_2D, numcomp, xpix, 
                ypix, GL_RGBA, GL_UNSIGNED_BYTE, TextureImage)!=0)
            {
                fprintf(stderr,"Mipmaps didn't build\n");
                exit(0);
            }
            break;
        }
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, texenvmode);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magfilter);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minfilter);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap);
        glEnable(GL_TEXTURE_2D);
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    }

    if(testnum)
    {
        st = testnum - 1;
        et = testnum;
    }
    else
    {
        st = 0;
        et = tc;
    }

    (*eventloop)(st,et,numframes);

    fprintf(stdout,"=============================================================\n");
}




